@@ -546,16 +546,14 @@ int sign (freenet_group *group, unsigned char *priv, unsigned char *pub,
 int verify (freenet_group *group, unsigned char *pub,
             unsigned char *buf, int len, unsigned char *sigstr)
 {
-  int status, retval, rlen, slen, i, j;
+  int status, retval, rlen, slen;
   int siglen;
   DSA *dsa;
   DSA_SIG *sig;
   unsigned char *comma;
   unsigned char *sigptr;
   unsigned char sigasn[FN_DHK_LEN];
-  unsigned char sigrbuf[FN_DHK_LEN];
-  unsigned char sigsbuf[FN_DHK_LEN];
-  unsigned char tmp[3];
+  unsigned char sigvalstr[(FN_DHK_LEN * 2) + 1];
 
   dsa = DSA_new();
   sig = DSA_SIG_new();
@@ -586,32 +584,33 @@ int verify (freenet_group *group, unsigned char *pub,
 
   comma = (unsigned char *)strchr(sigstr, ',');
   if (comma == NULL) {
-    return FNS_INVALID_SIG;
+    retval = FNS_INVALID_SIG;
+    goto end;
   }
 
-  slen = ((strlen(comma) - 1) + 1) / 2;
-  rlen = ((strlen(sigstr) - strlen(comma)) + 1) / 2;
+  rlen = (strlen(sigstr) - strlen(comma));
+  slen = (strlen(comma) - 1);
 
-  for (i = 0; i < rlen; i++) {
-    strncpy(tmp, &(sigstr[i*2]), 2);
-    tmp[2] = 0;
-    sigrbuf[i] = (unsigned char)strtol(tmp, NULL, 16);
+  if ((rlen > (FN_DHK_LEN*2)) ||
+      (slen > (FN_DHK_LEN*2))) {
+    retval = FNS_INVALID_SIG;
+    goto end;
   }
 
-  for (j = 0; j < slen; j++) {
-    strncpy(tmp, &(sigstr[(i*2) + 1 + (j*2)]), 2);
-    tmp[2] = 0;
-    sigsbuf[j] = (unsigned char)strtol(tmp, NULL, 16);
-  }
+  strncpy(sigvalstr, sigstr, rlen);
+  sigvalstr[rlen] = 0;
 
-  sig->r = BN_bin2bn(sigrbuf, rlen, NULL);
-  if (sig->r == NULL) {
+  status = BN_hex2bn(&(sig->r), sigvalstr);
+  if (status != strlen(sigvalstr)) {
     retval = FNS_BN_FAILED;
     goto end;
   }
 
-  sig->s = BN_bin2bn(sigsbuf, slen, NULL);
-  if (sig->s == NULL) {
+  strncpy(sigvalstr, &(comma[1]), slen);
+  sigvalstr[slen] = 0;
+
+  status = BN_hex2bn(&(sig->s), sigvalstr);
+  if (status != strlen(sigvalstr)) {
     retval = FNS_BN_FAILED;
     goto end;
   }
@@ -810,12 +809,11 @@ int stream_storedata (freenet_transfer_state *state)
   int status;
   freenet_message request;
 
-  strcpy(request.type, "StoreData");
+  request.type = FN_STOREDATA_MSG;
   request.uniqueid = state->uniqueid;
   strcpy(request.field[0], "Depth=0");
   sprintf(request.field[1], "HopsToLive=%x", state->htl);
-  sprintf(request.field[2], "EndMessage");
-  request.numfields=3;
+  request.numfields=2;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1013,7 +1011,7 @@ int freenet_request_stream (freenet_transfer_state *state, char *address,
     return status;
   }
 
-  strcpy(request.type, "DataRequest");
+  request.type = FN_DATAREQUEST_MSG;
   sprintf(request.field[0], "SearchKey=");
   for (i=0; i<FN_HASH_BYTES + 3; i++) {
     sprintf(&(request.field[0][10+(i*2)]), "%.2X", key->searchkey[i]);
@@ -1023,8 +1021,7 @@ int freenet_request_stream (freenet_transfer_state *state, char *address,
   sprintf(request.field[1], "Depth=%x", depth);
   sprintf(request.field[2], "HopsToLive=%x", htl);
   strcpy(request.field[3], "Source=tcp/localhost:6666");
-  strcpy(request.field[4], "EndMessage");
-  request.numfields = 5;
+  request.numfields = 4;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1043,23 +1040,19 @@ int freenet_request_stream (freenet_transfer_state *state, char *address,
     }
 
     i++;
-  } while ((strcmp(reply.type, "QueryRestarted") == 0) &&
+  } while ((reply.type == FN_QUERYRESTARTED_MSG) &&
            (i < FN_MAX_RESTARTS));
 
-  if (strcmp(reply.type, "DataReply") == 0) {
+  if (reply.type == FN_DATAREPLY_MSG) {
     len = 0;
 
-    if (strcmp(reply.field[reply.numfields-1], "Data")!=0) {
-      printf("trailing field on DataReply is %s\n",
-             reply.field[reply.numfields-1]);
-    }
-    for (i=0; i<reply.numfields-1; i++) {
+    for (i=0; i<reply.numfields; i++) {
       if (strncmp(reply.field[i], "Storable.Symmetric-cipher=", 26) == 0) {
         if (strcmp(reply.field[i],
-                    "Storable.Symmetric-cipher=Twofish") == 0) {
+                   "Storable.Symmetric-cipher=Twofish") == 0) {
           state->endtoend = FN_TWOFISH;
         } else if (strcmp(reply.field[i],
-                           "Storable.Symmetric-cipher=Rijndael") == 0) {
+                          "Storable.Symmetric-cipher=Rijndael") == 0) {
           state->endtoend = FN_RIJNDAEL;
         } else {
           return FNS_INVALID_ENDTOEND;
@@ -1164,9 +1157,9 @@ int freenet_request_stream (freenet_transfer_state *state, char *address,
 
     }
 
-  } else if (strcmp(reply.type, "TimedOut") == 0) {
+  } else if (reply.type == FN_TIMEDOUT_MSG) {
     return FNS_TIMEOUT;
-  } else if (strcmp(reply.type, "RequestFailed") == 0) {
+  } else if (reply.type == FN_REQUESTFAILED_MSG) {
     return FNS_REQUEST_FAILED;
   } else {
     return FNS_UNEXPECTED_MESSAGE;
@@ -1223,14 +1216,13 @@ int request_insert (freenet_transfer_state *state,
   }
   state->uniqueid = request.uniqueid;
 
-  strcpy(request.type, "InsertRequest");
+  request.type = FN_INSERTREQUEST_MSG;
   sprintf(request.field[0], "SearchKey=%s", searchkeystr);
   depth = (((float)rand() / (float)RAND_MAX) * (FN_DH - FN_DL)) + FN_DL;
   sprintf(request.field[1], "Depth=%x", depth);
   sprintf(request.field[2], "HopsToLive=%x", htl);
   strcpy(request.field[3], "Source=tcp/localhost:6666");
-  strcpy(request.field[4], "EndMessage");
-  request.numfields=5;
+  request.numfields=4;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1249,17 +1241,17 @@ int request_insert (freenet_transfer_state *state,
     }
 
     i++;
-  } while ((strcmp(reply.type, "QueryRestarted") == 0) &&
+  } while ((reply.type == FN_QUERYRESTARTED_MSG) &&
            (i < FN_MAX_RESTARTS));
 
 
-  if (strcmp(reply.type, "InsertReply")==0) {
+  if (reply.type == FN_INSERTREPLY_MSG) {
     return FNS_SUCCESS;
-  } else if (strcmp(reply.type, "TimedOut") == 0) {
+  } else if (reply.type == FN_TIMEDOUT_MSG) {
     return FNS_TIMEOUT;
-  } else if (strcmp(reply.type, "RequestFailed") == 0) {
+  } else if (reply.type == FN_REQUESTFAILED_MSG) {
     return FNS_REQUEST_FAILED;
-  } else if (strcmp(reply.type, "DataReply") == 0) {
+  } else if (reply.type == FN_DATAREPLY_MSG) {
     return FNS_DATA_REPLY;
   } else {
     return FNS_UNEXPECTED_MESSAGE;
@@ -1344,12 +1336,12 @@ int freenet_insert_CHK_buffer (freenet_transfer_state *state, char *address,
 
   request.uniqueid = state->uniqueid;
 
-  strcpy(request.type, "DataInsert");
+  request.type = FN_DATAINSERT_MSG;
   depth = (((float)rand() / (float)RAND_MAX) * (FN_DH - FN_DL)) + FN_DL;
   sprintf(request.field[0], "Depth=%x", depth);
   sprintf(request.field[1], "HopsToLive=%x", htl);
   sprintf(request.field[2], "DataLength=%x", total_len);
-  sprintf(request.field[3], "Storable.Metadata-length=%x", metadata_len);
+  sprintf(request.field[3], "Storable.Metadata-length=%d", metadata_len);
   if (current_part == 0) {
     strcpy(request.field[4], "Storable.PartSize=0");
   } else {
@@ -1360,8 +1352,7 @@ int freenet_insert_CHK_buffer (freenet_transfer_state *state, char *address,
   } else if (endtoend == FN_RIJNDAEL) {
     strcpy(request.field[5], "Storable.Symmetric-cipher=Rijndael");
   }
-  strcpy(request.field[6], "Data");
-  request.numfields=7;
+  request.numfields=6;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1490,12 +1481,12 @@ int freenet_insert_CHK_stream (freenet_transfer_state *state, char *address,
 
   request.uniqueid = state->uniqueid;
 
-  strcpy(request.type, "DataInsert");
+  request.type = FN_DATAINSERT_MSG;
   depth = (((float)rand() / (float)RAND_MAX) * (FN_DH - FN_DL)) + FN_DL;
   sprintf(request.field[0], "Depth=%x", depth);
   sprintf(request.field[1], "HopsToLive=%x", htl);
   sprintf(request.field[2], "DataLength=%x", total_len);
-  sprintf(request.field[3], "Storable.Metadata-length=%x", metadata_len);
+  sprintf(request.field[3], "Storable.Metadata-length=%d", metadata_len);
   if (current_part == 0) {
     strcpy(request.field[4], "Storable.PartSize=0");
   } else {
@@ -1506,8 +1497,7 @@ int freenet_insert_CHK_stream (freenet_transfer_state *state, char *address,
   } else if (endtoend == FN_RIJNDAEL) {
     strcpy(request.field[5], "Storable.Symmetric-cipher=Rijndael");
   }
-  strcpy(request.field[6], "Data");
-  request.numfields=7;
+  request.numfields=6;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1617,12 +1607,12 @@ int freenet_insert_SVK_buffer (freenet_transfer_state *state, char *address,
 
   request.uniqueid = state->uniqueid;
 
-  strcpy(request.type, "DataInsert");
+  request.type = FN_DATAINSERT_MSG;
   depth = (((float)rand() / (float)RAND_MAX) * (FN_DH - FN_DL)) + FN_DL;
   sprintf(request.field[0], "Depth=%x", depth);
   sprintf(request.field[1], "HopsToLive=%x", htl);
   sprintf(request.field[2], "DataLength=%x", total_len);
-  sprintf(request.field[3], "Storable.Metadata-length=%x", metadata_len);
+  sprintf(request.field[3], "Storable.Metadata-length=%d", metadata_len);
   strcpy(request.field[4], "Storable.PartSize=0");
   if (endtoend == FN_TWOFISH) {
     strcpy(request.field[5], "Storable.Symmetric-cipher=Twofish");
@@ -1642,8 +1632,7 @@ int freenet_insert_SVK_buffer (freenet_transfer_state *state, char *address,
   } else {
     fieldnum = 8;
   }
-  strcpy(request.field[fieldnum], "Data");
-  request.numfields=fieldnum+1;
+  request.numfields=fieldnum;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
@@ -1770,12 +1759,12 @@ int freenet_insert_SVK_stream (freenet_transfer_state *state, char *address,
 
   request.uniqueid = state->uniqueid;
 
-  strcpy(request.type, "DataInsert");
+  request.type = FN_DATAINSERT_MSG;
   depth = (((float)rand() / (float)RAND_MAX) * (FN_DH - FN_DL)) + FN_DL;
   sprintf(request.field[0], "Depth=%x", depth);
   sprintf(request.field[1], "HopsToLive=%x", htl);
   sprintf(request.field[2], "DataLength=%x", total_len);
-  sprintf(request.field[3], "Storable.Metadata-length=%x", metadata_len);
+  sprintf(request.field[3], "Storable.Metadata-length=%d", metadata_len);
   strcpy(request.field[4], "Storable.PartSize=0");
   if (endtoend == FN_TWOFISH) {
     strcpy(request.field[5], "Storable.Symmetric-cipher=Twofish");
@@ -1795,8 +1784,7 @@ int freenet_insert_SVK_stream (freenet_transfer_state *state, char *address,
   } else {
     fieldnum = 8;
   }
-  strcpy(request.field[fieldnum], "Data");
-  request.numfields=fieldnum+1;
+  request.numfields=fieldnum;
 
   status = freenet_sendmsg(&(state->connection), &request);
   if (status != FNS_SUCCESS) {
