@@ -385,12 +385,11 @@ int request_handshake (freenet_connection *connection)
     return status;
   }
 
-  strcpy(message.type, "HandshakeRequest");
+  message.type = FN_HANDSHAKEREQUEST_MSG;
   strcpy(message.field[0], "Depth=1");
   strcpy(message.field[1], "HopsToLive=1");
   strcpy(message.field[2], "KeepAlive=true");
-  strcpy(message.field[3], "EndMessage");
-  message.numfields = 4;
+  message.numfields = 3;
 
   status = freenet_sendmsg(connection, &message);
   if (status != FNS_SUCCESS) {
@@ -402,7 +401,7 @@ int request_handshake (freenet_connection *connection)
     return status;
   }
 
-  if (strcmp(reply.type, "HandshakeReply") != 0) {
+  if (reply.type != FN_HANDSHAKEREPLY_MSG) {
     return FNS_INVALID_REPLY;
   }
 
@@ -430,7 +429,7 @@ int reply_handshake (freenet_connection *connection, freenet_message *request)
   float version;
   int i, status;
 
-  if (strcmp(request->type, "HandshakeRequest") != 0) {
+  if (request->type != FN_HANDSHAKEREQUEST_MSG) {
     return FNS_INVALID_REQUEST;
   }
 
@@ -443,15 +442,14 @@ int reply_handshake (freenet_connection *connection, freenet_message *request)
     }
   }
 
-  strcpy(reply.type, "HandshakeReply");
+  reply.type = FN_HANDSHAKEREPLY_MSG;
   reply.uniqueid = request->uniqueid;
   sprintf(reply.field[0], "Version=%.3f", FN_VERSION);
   sprintf(reply.field[1], "Depth=1");
   sprintf(reply.field[2], "HopsToLive=1");
   sprintf(reply.field[3], "Revision=%.3f", FN_REVISION);
   sprintf(reply.field[4], "Build=%d", FN_BUILD);
-  sprintf(reply.field[5], "EndMessage");
-  reply.numfields = 6;
+  reply.numfields = 5;
 
   status = freenet_sendmsg(connection, &reply);
   if (status != FNS_SUCCESS) {
@@ -462,6 +460,68 @@ int reply_handshake (freenet_connection *connection, freenet_message *request)
 }
 
 
+int string_to_msgtype (const char *typestr)
+{
+  if (strcmp(typestr, "StoreData") == 0) {
+    return FN_STOREDATA_MSG;
+  } else if (strcmp(typestr, "HandshakeRequest") == 0) {
+    return FN_HANDSHAKEREQUEST_MSG;
+  } else if (strcmp(typestr, "HandshakeReply") == 0) {
+    return FN_HANDSHAKEREPLY_MSG;
+  } else if (strcmp(typestr, "QueryRestarted") == 0) {
+    return FN_QUERYRESTARTED_MSG;
+  } else if (strcmp(typestr, "DataRequest") == 0) {
+    return FN_DATAREQUEST_MSG;
+  } else if (strcmp(typestr, "DataReply") == 0) {
+    return FN_DATAREPLY_MSG;
+  } else if (strcmp(typestr, "TimedOut") == 0) {
+    return FN_TIMEDOUT_MSG;
+  } else if (strcmp(typestr, "RequestFailed") == 0) {
+    return FN_REQUESTFAILED_MSG;
+  } else if (strcmp(typestr, "InsertRequest") == 0) {
+    return FN_INSERTREQUEST_MSG;
+  } else if (strcmp(typestr, "InsertReply") == 0) {
+    return FN_INSERTREPLY_MSG;
+  } else if (strcmp(typestr, "DataInsert") == 0) {
+    return FN_DATAINSERT_MSG;
+  } else {
+    return FN_UNKNOWN_MSG;
+  }
+}
+
+
+int msgtype_to_string (int msgtype, char *typestr)
+{
+  if (msgtype == FN_STOREDATA_MSG) {
+    strcpy(typestr, "StoreData");
+  } else if (msgtype == FN_HANDSHAKEREQUEST_MSG) {
+    strcpy(typestr, "HandshakeRequest");
+  } else if (msgtype == FN_HANDSHAKEREPLY_MSG) {
+    strcpy(typestr, "HandshakeReply");
+  } else if (msgtype == FN_QUERYRESTARTED_MSG) {
+    strcpy(typestr, "QueryRestarted");
+  } else if (msgtype == FN_DATAREQUEST_MSG) {
+    strcpy(typestr, "DataRequest");
+  } else if (msgtype == FN_DATAREPLY_MSG) {
+    strcpy(typestr, "DataReply");
+  } else if (msgtype == FN_TIMEDOUT_MSG) {
+    strcpy(typestr, "TimedOut");
+  } else if (msgtype == FN_REQUESTFAILED_MSG) {
+    strcpy(typestr, "RequestFailed");
+  } else if (msgtype == FN_INSERTREQUEST_MSG) {
+    strcpy(typestr, "InsertRequest");
+  } else if (msgtype == FN_INSERTREPLY_MSG) {
+    strcpy(typestr, "InsertReply");
+  } else if (msgtype == FN_DATAINSERT_MSG) {
+    strcpy(typestr, "DataInsert");
+  } else {
+    return FNS_UNKNOWN_MSG_TYPE;
+  }
+
+  return FNS_SUCCESS;
+}
+
+
 int freenet_senddata(freenet_connection *connection, const char *buffer,
                      int len)
 {
@@ -536,24 +596,31 @@ int writeline(freenet_connection *connection, const char *string)
 int freenet_sendmsg (freenet_connection *connection, freenet_message *message)
 {
   int i, status;
-  unsigned char line[512];
+  unsigned char typeline[512];
+  unsigned char uniqline[512];
 
-  status = writeline(connection, message->type);
+
+  status = msgtype_to_string(message->type, typeline);
   if (status != FNS_SUCCESS) {
     return status;
   }
 
-  sprintf(line, "UniqueID=%Lx", message->uniqueid);
+  status = writeline(connection, typeline);
+  if (status != FNS_SUCCESS) {
+    return status;
+  }
 
-  status = writeline(connection, line);
+  sprintf(uniqline, "UniqueID=%Lx", message->uniqueid);
+
+  status = writeline(connection, uniqline);
   if (status != FNS_SUCCESS) {
     return status;
   }
 
   /* DEBUG */
   fprintf(stderr, "sending = \n");
-  fprintf(stderr, "%s\n", message->type);
-  fprintf(stderr, "%s\n", line);
+  fprintf(stderr, "%s\n", typeline);
+  fprintf(stderr, "%s\n", uniqline);
   for (i = 0; i < message->numfields; i++) {
     fprintf(stderr, "%s\n", message->field[i]);
   }
@@ -565,6 +632,19 @@ int freenet_sendmsg (freenet_connection *connection, freenet_message *message)
     }
   }
 
+  if ((message->type != FN_DATAREPLY_MSG) &&
+      (message->type != FN_DATAINSERT_MSG)) {
+    status = writeline(connection, "EndMessage");
+    if (status != FNS_SUCCESS) {
+      return status;
+    }
+  } else {
+    status = writeline(connection, "Data");
+    if (status != FNS_SUCCESS) {
+      return status;
+    }
+  }
+
   return FNS_SUCCESS;
 }
 
@@ -649,35 +729,51 @@ int readto(freenet_connection *connection, char *buffer, int buflen, int term)
 int internal_recvmsg (freenet_connection *connection, freenet_message *message)
 {
   int i, status;
+  unsigned char typeline[512];
   unsigned char line[512];
 
   message->numfields = 0;
-  status = readto(connection, message->type, 81, '\n');
+  status = readto(connection, typeline, 510, '\n');
   if (status == FNS_EOF) {
     return FNS_CONNECTION_GONE;
   } else if (status != FNS_SUCCESS) {
-    return FNS_SUCCESS;
+    return status;
   }
 
+  message->type = string_to_msgtype(typeline);
+
   do {
 
     status = readto(connection, line, 510, '\n');
     if (status != FNS_SUCCESS) {
-      return FNS_SUCCESS;
+      return status;
     }
 
     if (strncmp(line, "UniqueID=", 9) == 0) {
       sscanf(&(line[9]), "%Lx", &(message->uniqueid));
-    } else {
+    } else if (strchr(line, '=') != NULL) {
       strncpy(message->field[message->numfields], line, 510);
       message->numfields++;
     }
 
-  } while (strchr(message->field[message->numfields-1], '=') != NULL);
+  } while (strchr(line, '=') != NULL);
+
+
+  if ((message->type != FN_DATAREPLY_MSG) &&
+      (message->type != FN_DATAINSERT_MSG)) {
+    if (strcmp(line, "EndMessage") != 0) {
+      return FNS_BAD_TRAILING_FIELD;
+    }
+  } else {
+    if (strcmp(line, "Data") != 0) {
+      return FNS_BAD_TRAILING_FIELD;
+    }
+  }
+
 
   /* DEBUG */
   fprintf(stderr, "received = \n");
-  fprintf(stderr, "%s\n", message->type);
+  fprintf(stderr, "%s\n", typeline);
   fprintf(stderr, "UniqueID=%Lx\n", message->uniqueid);
   for (i = 0; i < message->numfields; i++) {
     fprintf(stderr, "%s\n", message->field[i]);
@@ -706,7 +802,7 @@ int freenet_recvmsg (freenet_connection *connection, freenet_message *message)
     return status;
   }
 
-  if (strcmp(message->type, "HandshakeRequest")==0) {
+  if (message->type == FN_HANDSHAKEREQUEST_MSG) {
     status = reply_handshake(connection, message);
     if (status != FNS_SUCCESS) {
       return status;
@@ -802,15 +898,3 @@ int freenet_get_connection (freenet_connection *connection, u_short port)
 
   return FNS_SUCCESS;
 }
-
-
-int freenet_init (void) {
-
-#ifndef FN_RANDOM_FILE
-  if (oc_start() != 0) {
-    return FNS_RANDOM_ERROR;
-  }
-#endif
-
-  return FNS_SUCCESS;
-}
